import os

from a2ml.api.utils import fsclient
from .deploy import ModelDeploy
from ..cloud.pipeline import AugerPipelineApi


class ModelUndeploy(object):
    """Undeploy Model locally or from Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelUndeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False):
        if locally:
            is_loaded, model_path, model_name = \
                ModelDeploy(self.ctx, self.project).verify_local_model(model_id)
            self.ctx.log("Undeploy model. Remove local model: %s" % model_name)

            if is_loaded:
                fsclient.remove_file(model_name)

            model_folder = os.path.splitext(model_name)[0]
            fsclient.remove_folder(model_folder)
        else:
            AugerPipelineApi(self.ctx, None).remove(model_id)



import json
import numpy as np
import math
import numbers


def is_nan(x):
    return (x is np.nan or x != x)

def convert_simple_numpy_type(obj):
    if isinstance(obj, (np.int_, np.intc, np.intp, np.int8,
        np.int16, np.int32, np.int64, np.uint8,
        np.uint16, np.uint32, np.uint64)):
        return int(obj)
    elif isinstance(obj, (np.float_, np.float16, np.float32,
        np.float64)):
        return float(obj)

    # elif isinstance(obj,(np.ndarray,)): #### This is the fix
    #     return obj.tolist()

    return None

class NumpyJSONEncoder(json.JSONEncoder):
    """ Special json encoder for numpy types """
    def default(self, obj):
        res = convert_simple_numpy_type(obj)
        if res is not None:
            return res

        return json.JSONEncoder.default(self, obj)

def convert_nan_inf(params):
    if type(params) is dict:
        for key, value in params.items():
            params[key] = convert_nan_inf(value)
    elif type(params) is list:
        for idx, value in enumerate(params):
            params[idx] = convert_nan_inf(value)
    else:
        if isinstance(params, numbers.Number) and math.isinf(params):
            params = None
        if is_nan(params):
            params = None

    return params

def json_dumps_np(data, allow_nan=False):
    if not allow_nan:
        data = convert_nan_inf(data)

    return json.dumps(data, cls=NumpyJSONEncoder, allow_nan=allow_nan)



from .cloud.dataset import AugerDataSetApi
from .exceptions import AugerException


class DataSet(AugerDataSetApi):
    """Auger Cloud Data Set(s) management"""

    def __init__(self, ctx, project, data_set_name=None):
        super(DataSet, self).__init__(
            ctx, project, data_set_name)
        self.project = project

    def create(self, data_source_file):
        if data_source_file is None:
            raise AugerException('Please specify data source file...')

        data_source_file, local_data_source = \
            AugerDataSetApi.verify(data_source_file, self.ctx.config.path)

        self.project.start()

        super().create(data_source_file, self.object_name, local_data_source=local_data_source)
        return self

    def upload_file(self, data_source_file):
        if data_source_file is None:
            raise AugerException('Please specify data source file...')

        data_source_file, local_data_source = \
            AugerDataSetApi.verify(data_source_file, self.ctx.config.path)

        self.project.start()

        return super().do_upload_file(data_source_file, self.object_name, local_data_source=local_data_source)

    def download(self, path_to_download):
        if path_to_download is None:
            raise AugerException('Please specify path to download...')

        self.project.start()

        return super().download(path_to_download)



import os
import shutil
import subprocess
from zipfile import ZipFile
import sys
import pandas as pd

from .deploy import ModelDeploy
from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..exceptions import AugerException
from a2ml.api.utils import fsclient
from a2ml.api.utils.dataframe import DataFrame
from a2ml.api.model_review.model_helper import ModelHelper
from ..decorators import with_project


class ModelPredict():
    """Predict using deployed Auger Model."""

    def __init__(self, ctx):
        super(ModelPredict, self).__init__()
        self.ctx = ctx

    def execute(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None):
        if filename and not (filename.startswith("http:") or filename.startswith("https:")) and\
            not fsclient.is_s3_path(filename):
            self.ctx.log('Predicting on data in %s' % filename)
            filename = os.path.abspath(filename)

        if locally:
            predicted = self._predict_locally(filename, model_id, threshold, data, columns, predicted_at, output)
        else:
            predicted = self._predict_on_cloud(filename, model_id, threshold, data, columns, predicted_at, output)

        return predicted

    @with_project(autocreate=False)
    def _upload_file_to_cloud(self, project, filename):
        from ..dataset import DataSet

        file_url, file_name = DataSet(self.ctx, project).upload_file(filename)
        return file_url

    def _process_input(self, filename, data, columns):
        send_records = False
        file_url = None
        records = None
        features = None
        max_records_size = 10*1024
        if filename:
            if not (filename.startswith("http:") or filename.startswith("https:")) and \
               not fsclient.is_s3_path(filename):
                file_size = fsclient.get_file_size(filename)
                if file_size <= max_records_size:
                    self.ctx.log("Local file '%s' with size %s(KB) is smaller %s KB, so send records to hub." % (filename, file_size/1024, max_records_size/1024))
                    send_records = True
            else:
                file_url = filename
                if fsclient.is_s3_path(file_url):
                    with fsclient.with_s3_downloaded_or_local_file(file_url) as local_path:
                        file_url = self._upload_file_to_cloud(local_path)
        elif data is not None and isinstance(data, pd.DataFrame):
            if len(data) < 100:
                send_records = True
        elif sys.getsizeof(data) > max_records_size:
            self.ctx.log("Size of data %s KB > %s KB, so send file to hub." % (sys.getsizeof(data)/1024, max_records_size/1024))
        else:
            send_records = True

        if send_records:
            ds = DataFrame.create_dataframe(filename, data, columns)
            records = ds.get_records()
            features = ds.columns
        else:
            if not file_url:
                if not filename:
                    with fsclient.save_atomic("temp_predict.csv.gz") as local_path:
                        ds = DataFrame.create_dataframe(filename, data, columns)
                        ds.saveToCsvFile(local_path)
                        file_url = self._upload_file_to_cloud(local_path)
                else:
                    file_url = self._upload_file_to_cloud(filename)

        return records, features, file_url
                    
    def _predict_on_cloud(self, filename, model_id, threshold, data, columns, predicted_at, output):
        records, features, file_url = self._process_input(filename, data, columns)
        pipeline_api = AugerPipelineApi(self.ctx, None, model_id)        
        predictions = pipeline_api.predict(records, features, threshold=threshold, file_url=file_url, predicted_at=predicted_at)            

        ds_result = DataFrame.create_dataframe(predictions.get('signed_prediction_url'),
            records=predictions.get('data'), features=predictions.get('columns'))
        temp_file = ds_result.options['data_path'] if predictions.get('signed_prediction_url') else None

        try:
            ds_result.options['data_path'] = filename
            ds_result.loaded_columns = columns
            return ModelHelper.save_prediction_result(ds_result,
                prediction_id = None, support_review_model = False,
                json_result=False, count_in_result=False, prediction_date=predicted_at,
                model_path=None, model_id=model_id, output=output)
        finally:
            if temp_file:
                fsclient.remove_file(temp_file)

    def _predict_locally(self, filename_arg, model_id, threshold, data, columns, predicted_at, output):
        model_deploy = ModelDeploy(self.ctx, None)
        is_model_loaded, model_path, model_name = \
            model_deploy.verify_local_model(model_id)

        if not is_model_loaded:
            raise AugerException('Model isn\'t loaded locally. '
                'Please use a2ml deploy command to download model.')

        model_path, model_existed = self._extract_model(model_name)
        model_options = fsclient.read_json_file(os.path.join(model_path, "model", "options.json"))

        filename = filename_arg
        if not filename:
            ds = DataFrame.create_dataframe(filename, data, columns)
            filename = os.path.join(self.ctx.config.get_path(), '.augerml', 'predict_data.csv')
            ds.saveToCsvFile(filename, compression=None)

        try:
            predicted = \
                self._docker_run_predict(filename, threshold, model_path)
        finally:
            # clean up unzipped model
            # if it wasn't unzipped before
            if not model_existed:
                fsclient.remove_folder(model_path)
                model_path = None

        if not filename_arg:
            ds_result = DataFrame.create_dataframe(predicted)

            ds_result.options['data_path'] = None
            ds_result.loaded_columns = columns

            return ModelHelper.save_prediction_result(ds_result,
                prediction_id = None,
                support_review_model = model_options.get("support_review_model") if model_path else False,
                json_result=False, count_in_result=False, prediction_date=predicted_at,
                model_path=model_path, model_id=model_id, output=output)
        elif output:
            fsclient.move_file(predicted, output)
            predicted = output

        return predicted

    def _extract_model(self, model_name):
        model_path = os.path.splitext(model_name)[0]
        model_existed = os.path.exists(model_path)

        if not model_existed:
            with ZipFile(model_name, 'r') as zip_file:
                zip_file.extractall(model_path)

        return model_path, model_existed

    def _docker_run_predict(self, filename, threshold, model_path):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')
        predict_file = os.path.basename(filename)
        data_path = os.path.abspath(os.path.dirname(filename))
        model_path = os.path.abspath(model_path)

        call_args = "--verbose=True --path_to_predict=./model_data/%s %s" % \
            (predict_file, "--threshold=%s" % str(threshold) if threshold else '')

        command = (r"docker run "
            "-v {model_path}:/var/src/auger-ml-worker/exported_model "
            "-v {data_path}:/var/src/auger-ml-worker/model_data "
            "deeplearninc/auger-ml-worker:{docker_tag} "
            "python ./exported_model/client.py {call_args}").format(
                model_path=model_path, data_path=data_path,
                docker_tag=docker_tag, call_args=call_args)

        try:
            self.ctx.log(
                'Running model in deeplearninc/'
                'auger-ml-worker:%s' % docker_tag)
            result_file = subprocess.check_output(command, stderr=subprocess.STDOUT, shell=True)
            result_file = result_file.decode("utf-8").strip()
            result_file = os.path.basename(result_file)
            # getattr(subprocess,
            #     'check_call' if self.ctx.debug else 'check_output')(
            #     command, stderr=subprocess.STDOUT, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Error running Docker container...')

        return os.path.join(data_path, "predictions", result_file)



from .base import AugerBaseApi
from .cluster_task import AugerClusterTaskApi


class AugerActualApi(AugerBaseApi):
    """Auger Trial API."""

    def __init__(self, ctx, pipeline_api, prediction_id=None):
        super(AugerActualApi, self).__init__(
            ctx, pipeline_api, prediction_id)

    def create(self, records, actuals_at, actuals_path=None):
        params = {
            'pipeline_id': self.parent_api.object_id,
        }

        if records:
            params['actuals'] = records

        if actuals_path:
            params['actuals_path'] = actuals_path

        if actuals_at:
            params['actuals_at'] = str(actuals_at)

        return self._call_create(
            params=params, 
            has_return_object=False)

    def delete(self, with_predictions, begin_date, end_date):
        res = self.rest_api.call( 'delete_%ss' % self.api_request_path, {
                'pipeline_id': self.parent_api.object_id,
                'with_predictions': with_predictions,
                'from': begin_date,
                'to': end_date
            })

        cluster_task = AugerClusterTaskApi(self.ctx, cluster_task_id=res['id'])
        cluster_task.wait_for_status(['pending', 'received', 'started', 'retry'])

        return True



from a2ml.api.auger.model import AugerModel
from a2ml.api.auger.dataset import AugerDataset
from a2ml.api.auger.experiment import AugerExperiment

class AugerA2ML(object):
    """Auger A2ML implementation."""

    def __init__(self, ctx):
        super(AugerA2ML, self).__init__()
        self.ctx = ctx

    def import_data(self, source=None):
        return AugerDataset(self.ctx).create(source=source)

    def train(self):
        return AugerExperiment(self.ctx).start()

    def evaluate(self, run_id = None):
        return AugerExperiment(self.ctx).leaderboard(run_id)

    def deploy(self, model_id, locally=False, review=True):
        return AugerModel(self.ctx).deploy(model_id, locally, review)

    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None):
        return AugerModel(self.ctx).predict(
            filename, model_id, threshold, locally, data, columns, predicted_at, output)

    def actuals(self, model_id, filename=None, actual_records=None, actuals_at=None, locally=False):
        return AugerModel(self.ctx).actuals(
            model_id, filename, actual_records, actuals_at, locally)

    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        return AugerModel(self.ctx).delete_actuals(
            model_id, with_predictions, begin_date, end_date, locally)



import arff
import copy
import json
import logging
import os
import pandas as pd
import warnings

from functools import wraps

from a2ml.api.utils import fsclient, get_uid, get_uid4, remove_dups_from_list, process_arff_line, download_file
from a2ml.api.utils.local_fsclient import LocalFSClient

# To avoid warnings for inplace operation on datasets
pd.options.mode.chained_assignment = None

class DataFrame(object):
    BOOLEAN_WORDS_TRUE = ['yes', 'on']
    BOOLEAN_WORDS_FALSE = ['no', 'off']

    def __init__(self, options):
        self.options = options
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]
        self.df = None
        self.dataset_name = None
        self.loaded_columns = None
        self.from_pandas = False

    def _get_compression(self, extension):
        compression = self.options.get('data_compression', 'infer')
        if extension.endswith('.gz') or extension.endswith('.gzip'):
            compression = 'gzip'
        elif extension.endswith('.bz2'):
            compression = 'bz2'
        elif extension.endswith('.zip'):
            compression = 'zip'
        elif extension.endswith('.xz'):
            compression = 'xz'

        return compression

    @staticmethod
    def create_dataframe(data_path=None, records=None, features=None):
        if data_path:
            ds = DataFrame({'data_path': data_path})
            ds.load(features = features)
        elif records is not None and isinstance(records, pd.DataFrame):
            ds = DataFrame({})
            ds.df = records
            ds.from_pandas = True
        else:    
            ds = DataFrame({})
            ds.load_records(records, features=features)

        return ds

    @staticmethod
    def load_from_files(files, features=None):
        for file in files:
            path = file if type(file) == str else file['path']
            df = DataFrame.create_dataframe(path, None, features)
            yield (file, df)

    def load_from_file(self, path, features=None, nrows=None):
        from collections import OrderedDict

        extension = path
        if self.options.get('data_extension', 'infer') != 'infer':
            extension = self.options['data_extension']

        if self.options.get('content_type') == 'multipart':
            fsclient.merge_folder_files(path)

        if extension.endswith('.arff') or extension.endswith('.arff.gz'):
            arffFile = None
            class ArffFile:
                def __init__(self, file):
                    self.file = file
                    self.date_attrs = {}

                def __iter__(self):
                    return self

                def __next__(self):
                    line = process_arff_line(next(self.file), self.date_attrs)
                    return line

            try:

                with fsclient.open(path, 'r') as f:
                    arffFile = ArffFile(f)
                    arff_data = arff.load(arffFile, return_type=arff.COO)

                convert_arff = DataFrame._convert_arff_coo
            except arff.BadLayout:
                with fsclient.open(path, 'r') as f:
                    arffFile = ArffFile(f)
                    arff_data = arff.load(arffFile, return_type=arff.DENSE)

                convert_arff = DataFrame._convert_arff_dense

            columns = [a[0] for a in arff_data['attributes']]
            series = convert_arff(features, columns, arff_data['data'])

            res = pd.DataFrame.from_dict(OrderedDict(
                (c, s) for c, s in zip(columns, series) if s is not None
            ))
            for date_field, fmt in arffFile.date_attrs.items():
                res[date_field] = pd.to_datetime(res[date_field], infer_datetime_format=True, errors='ignore', utc=True)

            return res
        elif extension.endswith('.pkl') or extension.endswith('.pkl.gz'):
            return self.loadFromBinFile(path, features)
        elif extension.endswith('.json') or extension.endswith('.json.gz'):
            path = fsclient.s3fs_open(path)
            return pd.read_json(path, orient=self.options.get('json_orient',None))
        elif extension.endswith('.xlsx') or extension.endswith('.xls'):
            path = fsclient.s3fs_open(path)
            return pd.read_excel(path)
        elif extension.endswith('.feather') or extension.endswith('.feather.gz') or extension.endswith('.feather.zstd') or extension.endswith('.feather.lz4'):
            # Features list is optional for feather file, but it can segfault without it on some files
            return self.loadFromFeatherFile(path, features)
        elif extension.endswith('.parquet'):
            return self.loadFromParquetFile(path)

        csv_with_header = self.options.get('csv_with_header', True)
        header = 0 if csv_with_header else None
        prefix = None if csv_with_header else 'c'

        compression = self._get_compression(extension)
        path = fsclient.s3fs_open(path)

        res_df = None
        try:
            res_df = pd.read_csv(
                path,
                encoding='utf-8',
                escapechar="\\",
                usecols=features,
                na_values=['?'],
                header=header,
                prefix=prefix,
                sep = ',',
                nrows=nrows,
                low_memory=False,
                compression=compression
            )
        except Exception as e:
            logging.error("read_csv failed: %s"%e)
            res_df = pd.read_csv(
                path,
                encoding='utf-8',
                escapechar="\\",
                usecols=features,
                na_values=['?'],
                header=header,
                prefix=prefix,
                sep = '|',
                nrows=nrows,
                low_memory=False,
                compression=compression
            )

        # if res_df is not None:
        #     for name, value in res_df.dtypes.items():
        #         if value == 'object':
        #             res_df[name] = pd.to_datetime(res_df[name], infer_datetime_format=True, errors='ignore', utc=True)

        return res_df

    def load(self, features=None, nrows=None):
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]

        import csv
        from io import StringIO

        path = self.options['data_path']
        if isinstance(path, StringIO):
            path.seek(0)
            self.df = pd.read_csv(path, encoding='utf-8', escapechar="\\", usecols=features, na_values=['?'], nrows=nrows)
            if self.options.get("targetFeature") in self.df.columns:
                self.dropna([self.options["targetFeature"]])
        else:
            if path.startswith("jdbc:"):
                import psycopg2
                from psycopg2.extensions import parse_dsn
                path = path.replace('sslfactory=org.postgresql.ssl.NonValidatingFactory&', '')
                ary = path.split('tablename')
                path = ary[0]
                tablename = ary[1]
                dataset_name = tablename

                self.dbconn_args = parse_dsn(path[5:])
                conn = psycopg2.connect(**self.dbconn_args)
                self.df = pd.read_sql("select * from %s"%tablename, con=conn)
            else:
                path, remote_path = self._check_remote_path()
                try:
                    self.df = self.load_from_file(path, features=features, nrows=nrows)
                except:
                    if remote_path:
                        logging.exception("Loading local file failed. Download it again...")
                        self.options['data_path'] = remote_path
                        path, remote_path = self._check_remote_path(force_download=True)
                        self.df = self.load_from_file(path, features=features, nrows=nrows)
                    else:
                        raise

                self.dataset_name = os.path.basename(path)

            if self.options.get("targetFeature") in self.df.columns:
                self.dropna([self.options["targetFeature"]])
        return self

    def _check_remote_path(self, force_download=False):
        remote_path = None
        if self.options['data_path'].startswith("http:") or self.options['data_path'].startswith("https:"):
            local_dir = LocalFSClient().get_temp_folder()
            file_name = 'data-' + get_uid4()

            local_file_path = download_file(self.options['data_path'],
                local_dir=local_dir, file_name=file_name, force_download=force_download)

            remote_path = self.options['data_path']
            self.options['data_path'] = local_file_path

        return self.options['data_path'], remote_path

    def load_records(self, records, features=None):
        self.categoricals = {}
        self.transforms_log = [[],[],[],[]]

        if features:
            self.df = pd.DataFrame.from_records(records, columns=features)
            self.loaded_columns = features
        else:
            self.df = pd.DataFrame(records) #dict

        return self

    def get_records(self):
        return self.df.values.tolist()

    def saveToCsvFile(self, path, compression="gzip"):
        fsclient.remove_file(path)
        fsclient.create_parent_folder(path)

        with fsclient.save_local(path) as local_path:
            self.df.to_csv(local_path, index=False, compression=compression, encoding='utf-8')

    def saveToBinFile(self, path):
        fsclient.save_object_to_file(self.df, path)

    def loadFromBinFile(self, path, features=None):
        self.df = fsclient.load_object_from_file(path)

        if features:
            self.df =  self.df[features]

        return self.df

    def saveToFeatherFile(self, path):
        fsclient.save_object_to_file(self.df, path, fmt="feather")

    def loadFromFeatherFile(self, path, features=None):
        self.df = fsclient.load_db_from_feather_file(path, features)
        return self.df

    def loadFromParquetFile(self, path, features=None):
        self.df = fsclient.load_db_from_parquet_file(path, features)    
        return self.df

    def saveToFile(self, path):
        if path.endswith('.feather') or path.endswith('.feather.gz') or path.endswith('.feather.zstd') or path.endswith('.feather.lz4'):
            self.saveToFeatherFile(path)
        else:
            compression = None
            if path.endswith('.gz'):
                compression = 'gzip'

            self.saveToCsvFile(path, compression)

    def count(self):
        if self.df is not None:
            return len(self.df)
        else:
            return 0

    @property
    def columns(self):
        return self.df.columns.tolist()

    def _map_dtypes(self, dtype):
        dtype_map = {'int64': 'integer', 'float64':'double', 'object': 'string',
            'categorical':'categorical', 'datetime64[ns]': 'datetime', 'bool': 'boolean'}
        if dtype_map.get(dtype, None):
            return dtype_map[dtype]

        if dtype and (dtype.startswith('int') or dtype.startswith('uint')):
            return 'integer'

        if dtype and dtype.startswith('float'):
            return 'float'

        if dtype and dtype.startswith('double'):
            return 'double'

        if dtype and dtype.startswith('datetime64'):
            return 'datetime'

        return dtype

    @property
    def dtypes(self):
        types_list = []
        columns_list = self.columns
        for idx, dtype in enumerate(self.df.dtypes):
            types_list.append((columns_list[idx], self._map_dtypes(dtype.name)))

        return types_list

    @property
    def dtypes_dict(self):
        types_dict = {}
        columns_list = self.columns
        for idx, dtype in enumerate(self.df.dtypes):
            types_dict[columns_list[idx]] = self._map_dtypes(dtype.name)

        return types_dict

    def select(self, features):
        self.df = self.df[features]
        return self

    def drop(self, columns):
        self.df.drop(columns, inplace=True, axis=1)

    def drop_duplicates(self, columns=None):
        self.df.drop_duplicates(subset=columns, inplace=True)
        self.df.reset_index(drop=True, inplace=True)
        return self

    def dropna(self, columns=None):
        self.df.dropna(subset=columns, inplace=True, axis=0)
        self.df.reset_index(drop=True, inplace=True)
        return self

    def fillna(self, value):
        if isinstance(value, dict):
            value = value.copy()
            for item in self.dtypes:
                if list(value.keys())[0] == item[0]:
                    if item[1] == 'string':
                        value[list(value.keys())[0]] = str(list(value.values())[0])
                    elif item[1] == 'integer':
                        value[list(value.keys())[0]] = int(list(value.values())[0])
                    else:
                        value[list(value.keys())[0]] = float(list(value.values())[0])

        self.df.fillna(value, inplace=True)
        return self

    def convertToCategorical(self, col_names, is_target = False, categories = None):
        #print("convertToCategorical:%s"%col_names)
        if not isinstance(col_names, list):
            col_names = [col_names]

        if is_target:
            for col in col_names:
                if col in self.columns and self.categoricals.get(col, None) is None:
                    self.df[col] = pd.Categorical(self.df[col], categories=categories)
                    self.categoricals[col] = {'categories': list(self.df[col].cat.categories)}
                    self.df[col] = self.df[col].cat.codes
        else:
            cols_to_process = []
            cols = self.columns
            for col in col_names:
                if col in cols:
                    cols_to_process.append(col)

            #print(cols_to_process)
            if cols_to_process:
                self.df = pd.get_dummies(self.df, columns=cols_to_process)
                new_cols = self.columns

                for col in cols_to_process:
                    generated_cols = []
                    for new_col in new_cols:
                        if new_col.startswith(col+'_'):
                            generated_cols.append(new_col)

                    self.categoricals[col] = {'columns': generated_cols}

        return self

    @staticmethod
    def _convert_arff_coo(features, columns, arff_data_data):
        if features is None:
            data = [([], []) for _ in columns]
        else:
            fset = remove_dups_from_list(features)
            data = [([], []) if c in fset else None for c in columns]

        for v, i, j in zip(*arff_data_data):
            d = data[j]
            if d is not None:
                indices, values = d
                if indices:
                    assert indices[-1] < i
                indices.append(i)
                values.append(v)

        max_i = -1
        for d in data:
            if d is not None and len(d[0]) > 0:
                max_i = max(max_i, d[0][-1])
        height = max_i + 1

        series = []
        for d in data:
            if d is None:
                s = None
            else:
                keys, values = d
                sa = pd.SparseArray(
                    values,
                    sparse_index=pd._libs.sparse.IntIndex(height, keys),
                    fill_value=0
                )
                s = pd.Series(sa.values)
            series.append(s)

        return series

    @staticmethod
    def _convert_arff_dense(features, columns, arff_data_data):
        if features is None or set(features) == set(columns):
            return zip(*arff_data_data)

        fset = remove_dups_from_list(features)
        return [
            [row[i] for row in arff_data_data] if c in fset else None
            for i, c in enumerate(columns)
        ]




import contextlib
import logging
import os
import shutil
import tempfile
import time

from .local_fsclient import LocalFSClient


def is_s3_path(path):
    return path.startswith("s3:/")


def _get_fsclient_bypath(path):
    if is_s3_path(path):
        from .s3_fsclient import S3FSClient
        return S3FSClient()
    else:
        return LocalFSClient()


def create_folder(path):
    client = _get_fsclient_bypath(path)
    client.create_folder(path)


def create_parent_folder(path):
    client = _get_fsclient_bypath(path)
    client.create_folder(get_parent_folder(path))


def get_parent_folder(path):
    return os.path.dirname(path)


def remove_folder(path, remove_self=True):
    client = _get_fsclient_bypath(path)
    client.remove_folder(path, remove_self)


def remove_file(path, wild=False):
    client = _get_fsclient_bypath(path)
    client.remove_file(path, wild)


def get_smart_open_transport_params(path):
    if is_s3_path(path):
        client = _get_fsclient_bypath(path)
        return client.get_smart_open_transport_params()

    return None


def open_file(path, mode, num_tries=20, encoding='utf-8', auto_decompression=True):
    import warnings

    if is_s3_path(path) and 'r' in mode:
        client = _get_fsclient_bypath(path)
        client.wait_for_path(path)

    with warnings.catch_warnings():
        warnings.simplefilter('ignore')
        import smart_open

        nTry = 0
        while nTry <= num_tries:
            try:
                # TODO: support append mode for s3
                return smart_open.open(
                    path,
                    mode,
                    encoding=encoding,
                    ignore_ext=not auto_decompression,
                    transport_params=get_smart_open_transport_params(path)
                )
            except Exception as e:
                if nTry >= num_tries:
                    raise

            if 'w' in mode:
                remove_file(path)

            nTry = nTry + 1
            time.sleep(1)


def s3fs_open(path, mode='rb'):
    if is_s3_path(path):
        client = _get_fsclient_bypath(path)
        return client.s3fs_open(path, mode)

    return path


def list_folder(path, wild=False, remove_folder_name=True, meta_info=False):
    client = _get_fsclient_bypath(path)

    if wild:
        if not client.is_folder_exists(get_parent_folder(path)):
            return []
    else:
        if not client.is_folder_exists(path):
            return []

    return client.list_folder(path, wild=wild, remove_folder_name=remove_folder_name, meta_info=meta_info)


def get_mtime(path):
    client = _get_fsclient_bypath(path)
    return client.get_mtime(path)


def get_file_size(path):
    client = _get_fsclient_bypath(path)
    return client.get_file_size(path)

# @classmethod
# def openFile(cls, path, mode):
#     client = cls._get_fsclient_bypath(path)
#     return client.open(path, mode)


def is_file_exists(path):
    client = _get_fsclient_bypath(path)
    return client.is_file_exists(path)


def is_folder_exists(path):
    client = _get_fsclient_bypath(path)
    return client.is_folder_exists(path)


def read_text_file(path):
    client = _get_fsclient_bypath(path)
    return client.read_text_file(path)


def write_text_file(path, data, atomic=False, mode="w"):
    client = _get_fsclient_bypath(path)
    client.write_text_file(path, data, atomic=atomic, mode=mode)


def write_json_file(path, data, atomic=False, allow_nan=False):
    from .json_utils import json_dumps_np
    write_text_file(path, json_dumps_np(
        data, allow_nan=allow_nan), atomic=atomic)


def update_json_file(path, data, atomic=False, allow_nan=False):
    from .json_utils import json_dumps_np
    fileData = read_json_file(path)
    fileData.update(data)
    write_text_file(path, json_dumps_np(
        fileData, allow_nan=allow_nan), atomic=atomic)


def read_json_file(path, check_if_exist=True, if_wait_for_file=False):
    import json

    if not is_s3_path(path):
        list_folder(get_parent_folder(path))

    wait_for_file(path, if_wait_for_file=if_wait_for_file)

    if check_if_exist and not is_file_exists(path):
        return {}

    nTry = 0
    while nTry < 10:
        json_text = read_text_file(path)
        try:
            return json.loads(json_text)
        except Exception as e:
            logging.error("Load json failed: %s.Text: %s" %
                          (repr(e), json_text))
            nTry += 1
            time.sleep(2)

    return {}


def wait_for_file(path, if_wait_for_file, num_tries=30, interval_sec=1):
    if if_wait_for_file:
        nTry = 0
        while nTry <= num_tries:
            if is_file_exists(path):
                break

            if path != "/mnt/ready.txt":
                wait_for_fs_ready()

            if is_file_exists(path):
                break

            logging.info("File %s does not exist. Wait for %s sec" %
                         (path, interval_sec))

            nTry = nTry + 1
            time.sleep(interval_sec)

        return is_file_exists(path)

    return True


def wait_for_fs_ready():
    if not os.environ.get('S3_DATA_PATH') and os.environ.get('AUGER_ROOT_DIR', '').startswith('/mnt'):
        return wait_for_file("/mnt/ready.txt", if_wait_for_file=True, num_tries=30, interval_sec=10)

    return True


def load_json_files(paths):
    result = []
    for path in paths:
        try:
            result.append(read_json_file(path))
        except Exception as e:
            logging.exception("loadJSONFiles failed.")

    return result


def copy_file(path_src, path_dst, check_if_exist=True):

    if check_if_exist:
        client = _get_fsclient_bypath(path_src)
        if not client.is_file_exists(path_src):
            return

    client = _get_fsclient_bypath(path_dst)
    client.copy_file(path_src, path_dst)


def copy_files(path_src, path_dst):
    client = _get_fsclient_bypath(path_src)
    client.copy_files(path_src, path_dst)


def archive_folder(path_src, fmt='zip'):
    if is_s3_path(path_src):
        localClient = LocalFSClient()

        with localClient.save_atomic(path_src) as local_path:
            logging.info("archiveDir local path:%s" % local_path)

            clientS3 = _get_fsclient_bypath(path_src)
            clientS3.download_folder(path_src, local_path)

            localClient.archive_folder(local_path, fmt)

            clientS3.move_file(local_path+'.zip', path_src+'.zip')
            return

    client = _get_fsclient_bypath(path_src)
    client.archive_folder(path_src, fmt)


def copy_folder(path_src, path_dst):
    if is_folder_exists(path_dst):
        remove_folder(path_dst)

    if is_s3_path(path_dst):
        client = _get_fsclient_bypath(path_dst)
        client.copy_folder(path_src, path_dst)
    else:
        client = _get_fsclient_bypath(path_src)
        client.copy_folder(path_src, path_dst)


def _save_to_pickle(obj, path, compress):
    import joblib

    joblib.dump(obj, path, compress=compress, protocol=4)

def _save_to_feather(obj, path, compress):
    from pyarrow import feather

    feather.write_feather(obj, path, compression=compress)

def save_object_to_file(obj, path, fmt="pickle"):
    
    remove_file(path)
    create_parent_folder(path)

    try:
        compress = 0
        if path.endswith('.gz'):
            compress = ('gzip', 3)
        elif path.endswith('.zstd'):
            compress = "zstd"
        elif path.endswith('.lz4'):
            compress = "lz4"

        if is_s3_path(path):
            with save_atomic(path) as local_path:
                if fmt == "pickle":
                    _save_to_pickle(obj, local_path, compress=compress)
                elif fmt == "feather":
                    _save_to_feather(obj, local_path, compress=compress)
        else:
            if fmt == "pickle":
                _save_to_pickle(obj, path, compress=compress)
            elif fmt == "feather":
                _save_to_feather(obj, path, compress=compress)

    except:
        remove_file(path)
        raise


@contextlib.contextmanager
def save_atomic(path, move_file=True):
    localClient = LocalFSClient()

    with localClient.save_atomic(path) as local_path:
        yield local_path

        if move_file:
            client = _get_fsclient_bypath(path)
            client.move_file(local_path, path)


@contextlib.contextmanager
def save_local(path, move_file=True):
    if is_s3_path(path):
        localClient = LocalFSClient()

        with localClient.save_atomic(path) as local_path:
            yield local_path

            if move_file:
                client = _get_fsclient_bypath(path)
                client.move_file(local_path, path)
    else:
        yield path


@contextlib.contextmanager
def with_cur_dir(dir_path):
    create_folder(dir_path)

    old_path = os.getcwd()
    os.chdir(dir_path)
    yield
    os.chdir(old_path)

def move_file(path_src, path_dst):
    #print(path_src, path_dst)

    client = _get_fsclient_bypath(path_dst)
    client.move_file(path_src, path_dst)


def download_file(path, local_path):
    if path.startswith("http:") or path.startswith("https:"):
        client = _get_fsclient_bypath(local_path)
        client.download_file(path, local_path)
    elif is_s3_path(path):
        client = _get_fsclient_bypath(path)
        client.download_file(path, local_path)
    else:
        copy_file(path, local_path)

@contextlib.contextmanager
def with_s3_downloaded_or_local_file(source_path):
    if source_path.startswith("s3:"):
        tmp_dir = tempfile.mkdtemp()
        try:
            tmp_file = os.path.join(tmp_dir, os.path.basename(source_path))
            download_file(source_path, tmp_file)
            yield tmp_file
        finally:
            shutil.rmtree(tmp_dir)
    else:
        yield source_path


def load_object_from_file(path, use_local_cache=False):
    import joblib

    path_to_load = None
    if is_s3_path(path):
        if use_local_cache:
            local_path = path.replace(
                "s3://"+os.environ.get('S3_DATA_PATH'), os.environ.get("AUGER_LOCAL_TMP_DIR", ''))
            #logging.info("Local cache path: %s"%local_path)
            if not is_file_exists(local_path):
                local_lock_path = local_path + '.lock'
                create_parent_folder(local_lock_path)
                f_lock = None
                try:
                    f_lock = open(local_lock_path, 'x')
                except Exception as e:
                    #logging.exception("Open lock file failed.")
                    pass

                if f_lock:
                    try:
                        if not is_file_exists(local_path):
                            with save_atomic(local_path) as local_tmp_path:
                                logging.info("Download file from s3 to: %s, temp folder: %s" % (
                                    local_path, local_tmp_path))
                                download_file(path, local_tmp_path)
                    finally:
                        f_lock.close()
                        remove_file(local_lock_path)
                else:
                    wait_for_file(local_path, True,
                                  num_tries=300, interval_sec=10)

            path_to_load = local_path
        else:
            with save_atomic(path, move_file=False) as local_tmp_path:
                download_file(path, local_tmp_path)
                return joblib.load(local_tmp_path)
    else:
        path_to_load = path

    return joblib.load(path_to_load)

def load_npobject_from_file(path):
    import numpy as np

    if is_s3_path(path):
        with save_atomic(path, move_file=False) as local_path:
            download_file(path, local_path)
            return np.load(local_path, allow_pickle=True)

    return np.load(path, allow_pickle=True)

def _read_feather(path, features=None):
    from pyarrow import feather

    if path.endswith(".gz") or path.endswith(".zip"):
        with open_file(path, 'rb', encoding=None) as local_file:
            return feather.read_feather(local_file, columns=features, use_threads=bool(True))

    return feather.read_feather(path, columns=features, use_threads=bool(True))

def load_db_from_feather_file(path, features=None):
    if is_s3_path(path):
        with save_atomic(path, move_file=False) as local_path:
            download_file(path, local_path)
            return _read_feather(local_path, features)

    return _read_feather(path, features)

def _read_parquet(path, features=None):
    import pandas as pd

    return pd.read_parquet(path, columns=features)

def load_db_from_parquet_file(path, features=None):
    if is_s3_path(path):
        with save_atomic(path, move_file=False) as local_path:
            download_file(path, local_path)
            return _read_parquet(local_path, features)

    return _read_parquet(path, features)

def is_abs_path(path):
    import platform

    if not path:
        return False

    if is_s3_path(path):
        return True

    if path.startswith("http:") or path.startswith("https:"):
        return True

    if platform.system().lower() == 'windows':
        return ':\\' in path or ':/' in path

    return path[0] == "/" or path[0] == "\\"


def merge_folder_files(path, remove_folder=True):
    if is_file_exists(path):
        return

    res = os.path.splitext(path)
    path_no_ext = res[0]
    extension = res[1]
    if not is_folder_exists(path_no_ext):
        return

    client = _get_fsclient_bypath(path)
    files = list(client.list_folder(path_no_ext))

    with open_file(path, 'w') as file_output:
        for idx, file in enumerate(files):
            if file.endswith(extension):
                with open_file(os.path.join(path_no_ext, file), 'rb') as file_input:
                    data_file = file_input.read()
                    if extension == ".json":
                        data_file = _process_merged_json(
                            data_file, idx, len(files))

                    file_output.write(data_file)

    if remove_folder:
        client.remove_folder(path_no_ext)


def _process_merged_json(data_file, idx, total_len):
    remove_last = False
    remove_first = False

    if idx == 0:
        remove_last = True
    elif idx == total_len-1:
        remove_first = True
    else:
        remove_last = True
        remove_first = True

    if remove_last:
        index = data_file.rindex(']')
        if index >= 0:
            data_file = data_file[:index] + "," + data_file[index + 1:]
    if remove_first:
        index = data_file.index('[')
        if index >= 0:
            data_file = data_file[:index] + data_file[index + 1:]

    return data_file



import os
import pytest

from .utils import interceptor, object_status_chain, ORGANIZATIONS, PROJECTS
from a2ml.api.auger.model import AugerModel
from a2ml.api.utils import fsclient


class TestModel():
    def test_deploy_locally(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = {
            'get_organizations': ORGANIZATIONS,
            'get_projects': PROJECTS,
        }
        interceptor(PAYLOAD, monkeypatch)
        object_status_chain(['undeployed', 'deployed', 'deploying', 'running'], monkeypatch)
        monkeypatch.setattr('a2ml.api.auger.impl.mparts.deploy.ModelDeploy._docker_pull_image', lambda self: 'experimental')
        monkeypatch.setattr('a2ml.api.auger.impl.cloud.project.AugerProjectApi.start', lambda self: None)
        os.remove('models/model-87C81FE615DE46D.zip')
        # FIXME: let AugerPipelineFileApi do it's work
        monkeypatch.setattr('a2ml.api.auger.impl.cloud.pipeline_file.AugerPipelineFileApi.create', lambda self, model_id: {'signed_s3_model_path': 'None'})
        monkeypatch.setattr('a2ml.api.auger.impl.cloud.pipeline_file.AugerPipelineFileApi.download', lambda *a, **kw: 'models/export-%s.zip' % '87C81FE615DE46D')

        result = AugerModel(ctx).deploy('87C81FE615DE46D', locally=True, review=True)
        assert result['model_id'] == '87C81FE615DE46D'

    def test_undeploy_locally(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = {
            'get_organizations': ORGANIZATIONS,
            'get_projects': PROJECTS,
        }
        interceptor(PAYLOAD, monkeypatch)

        model_folder = "models/model-87C81FE615DE46D"
        model_file = "models/model-87C81FE615DE46D.zip"
        fsclient.create_folder(model_folder)
        assert fsclient.is_folder_exists(model_folder)
        fsclient.write_text_file(model_file, "TEST")
        assert fsclient.is_file_exists(model_file)

        result = AugerModel(ctx).undeploy('87C81FE615DE46D', locally=True)
        assert result['model_id'] == '87C81FE615DE46D'
        assert not fsclient.is_folder_exists(model_folder)
        assert not fsclient.is_file_exists(model_file)

    @pytest.mark.skip(reason="not implemented on server-side currently")
    def test_deploy_remoteley(self, log, project, ctx, authenticated, monkeypatch):
        result = runner.invoke(cli, ['model', 'deploy'])
        pass

    def test_predict_locally(self, log, project, ctx, authenticated, monkeypatch):
        PAYLOAD = {
            'get_organizations': ORGANIZATIONS,
            'get_projects': PROJECTS,
        }
        interceptor(PAYLOAD, monkeypatch)
        monkeypatch.setattr('subprocess.check_output', lambda *a, **kw: b'iris_predicted.csv')
        monkeypatch.setattr('subprocess.check_call', lambda *a, **kw: b'iris_predicted.csv')

        result = AugerModel(ctx).predict(filename='iris.csv', model_id='87C81FE615DE46D', 
            threshold=None, locally=True, data=None, columns=None, predicted_at=None, output=None)

        assert result.get('predicted')
        assert 'test_project/predictions/iris_predicted.csv' in result.get('predicted')


    @pytest.mark.skip(reason="not implemented on server-side currently")
    def test_predict_remoteley(self, log, project, ctx, authenticated, monkeypatch):
        result = runner.invoke(cli, ['model', 'predict', 'iris.csv'])
        pass

    @pytest.mark.skip(reason="not implemented on server-side currently")
    def test_actual_remoteley(self, log, project, ctx, authenticated, monkeypatch):
        result = runner.invoke(cli, ['model', 'actuals', 'iris.csv'])
        pass



import os
import subprocess

from ..cloud.cluster import AugerClusterApi
from ..cloud.pipeline import AugerPipelineApi
from ..exceptions import AugerException
from ..cloud.pipeline_file import AugerPipelineFileApi
from a2ml.api.utils import fsclient

class ModelDeploy(object):
    """Deploy Model on locally or on Auger Cloud."""

    def __init__(self, ctx, project):
        super(ModelDeploy, self).__init__()
        self.project = project
        self.ctx = ctx

    def execute(self, model_id, locally=False, review=True):
        if locally:
            return self.deploy_model_locally(model_id, review)
        else:
            return self.deploy_model_in_cloud(model_id, review)

    def deploy_model_in_cloud(self, model_id, review):
        self.ctx.log('Deploying model %s' % model_id)

        self.project.start()

        pipeline_properties = AugerPipelineApi(
            self.ctx, None).create(model_id, review)

        self.ctx.log('Deployed Model on Auger Cloud. Model id is %s' % \
            pipeline_properties.get('id'))

        return pipeline_properties.get('id')

    def deploy_model_locally(self, model_id, review):
        is_loaded, model_path, model_name = self.verify_local_model(model_id)
        #TODO: support review flag
        if not is_loaded:
            self.ctx.log('Downloading model %s' % model_id)

            self.project.start()

            pipeline_file_api = AugerPipelineFileApi(self.ctx, None)
            pipeline_file_properties = pipeline_file_api.create(model_id)
            downloaded_model_file = pipeline_file_api.download(
                pipeline_file_properties['signed_s3_model_path'],
                model_path, model_id)

            self.ctx.log('Downloaded model to %s' % downloaded_model_file)

            self.ctx.log('Pulling docker image required to predict')
            self._docker_pull_image()            
        else:
            self.ctx.log('Downloaded model is %s' % model_name)

        return model_id

    def verify_local_model(self, model_id):
        model_path = os.path.join(self.ctx.config.get_path(), 'models')
        model_name = os.path.join(model_path, 'model-%s.zip' % model_id)
        return fsclient.is_file_exists(model_name), model_path, model_name

    def _docker_pull_image(self):
        cluster_settings = AugerClusterApi.get_cluster_settings(self.ctx)
        docker_tag = cluster_settings.get('kubernetes_stack')

        try:
            subprocess.check_call(
                'docker pull deeplearninc/auger-ml-worker:%s' % \
                 docker_tag, shell=True)
        except subprocess.CalledProcessError as e:
            raise AugerException('Can\'t pull Docker container...')

        return docker_tag



import os

from .mparts.deploy import ModelDeploy
from .mparts.undeploy import ModelUndeploy
from .mparts.predict import ModelPredict
from .mparts.actual import ModelActual
from .mparts.delete_actual import ModelDeleteActual
from a2ml.api.model_review.model_review import ModelReview


class Model(object):
    """Auger Cloud Model(s) management."""

    def __init__(self, ctx, project):
        super(Model, self).__init__()
        self.project = project
        self.ctx = ctx

    def deploy(self, model_id, locally=False, review=True):
        return ModelDeploy(self.ctx, self.project).execute(model_id, locally, review)

    def undeploy(self, model_id, locally=False):
        return ModelUndeploy(self.ctx, self.project).execute(model_id, locally)

    def predict(self, filename, model_id, threshold=None, locally=False, data=None, columns=None, predicted_at=None, output=None):
        if locally:
            self.deploy(model_id, locally)
            
        return ModelPredict(self.ctx).execute(filename, model_id, threshold, locally, data, columns, predicted_at, output)

    def actuals(self, model_id, filename=None, actual_records=None, actuals_at=None, locally=False):
        if locally:
            is_loaded, model_path, model_name = ModelDeploy(self.ctx, self.project).\
                verify_local_model(model_id)

            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            model_path, model_existed = ModelPredict(self.ctx)._extract_model(model_name)
            return ModelReview({'model_path': os.path.join(model_path, "model")}).add_actuals(
              actuals_path=filename, actual_records=actual_records, actual_date=actuals_at)
        else:    
            return ModelActual(self.ctx).execute(model_id, filename, actual_records, actuals_at)

    def delete_actuals(self, model_id, with_predictions=False, begin_date=None, end_date=None, locally=False):
        if locally:
            is_loaded, model_path, model_name = ModelDeploy(self.ctx, self.project).\
                verify_local_model(model_id)

            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            model_path, model_existed = ModelPredict(self.ctx)._extract_model(model_name)
            return ModelReview({'model_path': os.path.join(model_path, "model")}).delete_actuals(
              with_predictions=with_predictions, begin_date=begin_date, end_date=end_date)
        else:    
            return ModelDeleteActual(self.ctx).execute(model_id, with_predictions, begin_date, end_date)

    def build_review_data(self, model_id, locally, output):
        if locally:
            is_loaded, model_path, model_name = ModelDeploy(self.ctx, self.project).\
                verify_local_model(model_id)

            if not is_loaded:
                raise AugerException('Model should be deployed locally.')

            model_path, model_existed = ModelPredict(self.ctx)._extract_model(model_name)
            return ModelReview({'model_path': os.path.join(model_path, "model")}).build_review_data(
              data_path=self.ctx.config.get("source"), output=output)
        else:
            raise Exception("Not Implemented.")
